"""
MovementControls class.
"""

# └── functions/movements.py
#     ├── [BROKER] get_xyz()
#     ├── [BROKER] move()
#     ├── [BROKER] set_home()
#     ├── [BROKER] find_home()
#     ├── [BROKER] find_axis_length()
#     └── [BROKER] check_position()

from .broker import BrokerConnect
from .information import Information

AXES = ["x", "y", "z", "all"]


def validate_axis(axis):
    """Validate axis."""
    if axis not in AXES:
        raise ValueError(f"Invalid axis: {axis} not in {AXES}")


class MovementControls():
    """MovementControls class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.info = Information(state)
        self.state = state

    def move(self, x=None, y=None, z=None, safe_z=None, speed=None):
        """Moves to the specified (x, y, z) coordinate."""
        self.state.print_status(description=f"Moving to ({x}, {y}, {z}).")

        def axis_overwrite(axis, value):
            return {
                "kind": "axis_overwrite",
                "args": {
                    "axis": axis,
                    "axis_operand": {
                        "kind": "numeric",
                        "args": {
                            "number": value
                        }
                    }
                }
            }

        def safe_z_body_item():
            return {
                "kind": "safe_z",
                "args": {},
            }

        def speed_overwrite(axis, speed):
            return {
                "kind": "speed_overwrite",
                "args": {
                    "axis": axis,
                    "speed_setting": {
                        "kind": "numeric",
                        "args": {
                            "number": speed,
                        }
                    }
                }
            }

        move_message = {
            "kind": "move",
            "args": {},
            "body": [],
        }

        if x is not None:
            move_message["body"].append(axis_overwrite("x", x))

        if y is not None:
            move_message["body"].append(axis_overwrite("y", y))

        if z is not None:
            move_message["body"].append(axis_overwrite("z", z))

        if speed is not None:
            move_message["body"].append(speed_overwrite("x", speed))
            move_message["body"].append(speed_overwrite("y", speed))
            move_message["body"].append(speed_overwrite("z", speed))

        if safe_z is not None:
            move_message["body"].append(safe_z_body_item())

        self.broker.publish(move_message)

    def set_home(self, axis="all"):
        """Sets the current position as the home position for a specific axis."""
        self.state.print_status(description="Setting home position")

        validate_axis(axis)

        set_home_message = {
            "kind": "zero",
            "args": {
                "axis": axis
            }
        }
        self.broker.publish(set_home_message)

    def find_home(self, axis="all", speed=100):
        """Moves the device to the home position for a specified axis."""
        self.state.print_status(description="Finding home position")

        validate_axis(axis)

        if speed > 100 or speed < 1:
            error = "ERROR: Speed constrained to 1-100."
            self.state.print_status(description=error, update_only=True)
            self.state.error = error
            return

        message = {
            "kind": "find_home",
            "args": {
                "axis": axis,
                "speed": speed
            }
        }
        self.broker.publish(message)

    def find_axis_length(self, axis="all"):
        """Finds the length of a specified axis."""
        self.state.print_status(description="Finding axis length")

        validate_axis(axis)

        find_axis_length_message = {
            "kind": "calibrate",
            "args": {
                "axis": axis
            }
        }

        self.broker.publish(find_axis_length_message)

    def get_xyz(self):
        """Returns the current (x, y, z) coordinates of the FarmBot."""
        self.state.print_status(description="Getting current coordinates")

        tree_data = self.info.read_status()
        if tree_data is None:
            error = "ERROR: No location data available."
            self.state.print_status(description=error, update_only=True)
            self.state.error = error
            return None
        position = tree_data["location_data"]["position"]

        self.state.print_status(
            description=f"Current position: {position}.",
            update_only=True)
        return position

    def check_position(self, coordinate, tolerance):
        """Verifies position of the FarmBot within specified tolerance range."""

        self.state.print_status(
            description=f"Checking if position is {coordinate} with tolerance: {tolerance}.")

        actual_vals = self.get_xyz()

        if actual_vals is None:
            return False

        for axis in ['x', 'y', 'z']:
            user_value = coordinate[axis]
            actual_value = actual_vals[axis]
            if not actual_value - tolerance <= user_value <= actual_value + tolerance:
                description = "Farmbot is NOT at position."
                description += f"\n Current position: {actual_vals}."
                self.state.print_status(
                    description=description,
                    update_only=True)
                return False

        self.state.print_status(
            description=f"Farmbot is at position: {actual_vals}.",
            update_only=True)
        return True



"""
MessageHandling class.
"""

# └── functions/messages.py
#     ├── [API] log()
#     ├── [BROKER] message()
#     ├── [BROKER] debug()
#     └── [BROKER] toast()

from .broker import BrokerConnect
from .api import ApiConnect
from .information import Information

MESSAGE_TYPES = ["assertion", "busy", "debug",
                 "error", "fun", "info", "success", "warn"]
CHANNELS = ["ticker", "toast", "email", "espeak"]


def validate_log_options(message_type, channels):
    """Validate the message type and channel options."""
    if message_type not in MESSAGE_TYPES:
        raise ValueError(
            f"Invalid message type: `{message_type}` not in {MESSAGE_TYPES}")
    for channel in channels:
        if channel not in CHANNELS:
            raise ValueError(f"Invalid channel: {channel} not in {CHANNELS}")


class MessageHandling():
    """Message handling class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.api = ApiConnect(state)
        self.info = Information(state)
        self.state = state

    def log(self, message_str, message_type="info", channels=None):
        """Sends new log message via the API."""
        self.state.print_status(
            description="Sending new log message to the API.")

        channels = channels or []
        validate_log_options(message_type, channels)

        log_message = {
            "message": message_str,
            "type": message_type,
            "channels": channels,
        }

        self.info.api_post("logs", log_message)

    def send_message(self, message_str, message_type="info", channels=None):
        """Sends new log message via the message broker."""
        self.state.print_status(
            description="Sending new log message to the message broker.")

        channels = channels or []
        validate_log_options(message_type, channels)

        message = {
            "kind": "send_message",
            "args": {
                "message": message_str,
                "message_type": message_type,
            },
            "body": [],
        }

        for channel in channels:
            message["body"].append({
                "kind": "channel",
                "args": {"channel_name": channel}})

        self.broker.publish(message)

    def debug(self, message_str):
        """Sends debug message used for developer information or troubleshooting."""
        self.state.print_status(
            description="Sending debug message to the message broker.")

        self.send_message(message_str, "debug")

    def toast(self, message_str, message_type="info"):
        """Sends a message that pops up on the user interface briefly."""
        self.state.print_status(
            description="Sending toast message to the message broker.")

        self.send_message(message_str, message_type, channels=["toast"])



"""
Peripherals class.
"""

# └── functions/peripherals.py
#     ├── [BROKER] control_servo()
#     ├── [BROKER] write_pin()
#     ├── [BROKER] control_peripheral()
#     ├── [BROKER] toggle_peripheral()
#     ├── [BROKER] on()
#     └── [BROKER] off()

from .broker import BrokerConnect
from .information import Information


class Peripherals():
    """Peripherals class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.info = Information(state)
        self.state = state

    def control_servo(self, pin, angle):
        """Set servo angle between 0-180 degrees."""
        self.state.print_status(description=f"Setting servo angle to {angle}.")

        if angle < 0 or angle > 180:
            error = "ERROR: Servo angle constrained to 0-180 degrees."
            self.state.print_status(description=error, update_only=True)
            self.state.error = error
            return

        control_servo_message = {
            "kind": "set_servo_angle",
            "args": {
                "pin_number": pin,
                "pin_value": angle  # From 0 to 180
            }
        }

        self.broker.publish(control_servo_message)

        return

    def write_pin(self, pin_number, value, mode="digital"):
        """Write a value to a pin."""
        pin_mode = self.info.convert_mode_to_number(mode)
        self.state.print_status(
            description=f"Setting pin {pin_number} to {value} ({mode}).")

        write_pin_message = {
            "kind": "write_pin",
            "args": {
                "pin_number": pin_number,
                "pin_value": value,
                "pin_mode": pin_mode,
            }
        }

        self.broker.publish(write_pin_message)

    def control_peripheral(self, peripheral_name, value, mode=None):
        """Set peripheral value and mode."""
        self.state.print_status(
            description=f"Setting {peripheral_name} to {value}.")

        peripheral = self.info.get_resource_by_name(
            "peripherals",
            peripheral_name)
        if peripheral is None:
            return
        peripheral_id = peripheral["id"]
        pin_mode = peripheral["mode"]
        if mode is not None:
            pin_mode = self.info.convert_mode_to_number(mode)

        control_peripheral_message = {
            "kind": "write_pin",
            "args": {
                "pin_value": value,
                "pin_mode": pin_mode,
                "pin_number": {
                    "kind": "named_pin",
                    "args": {
                        "pin_type": "Peripheral",
                        "pin_id": peripheral_id,
                    }
                }
            }
        }

        self.broker.publish(control_peripheral_message)

    def toggle_peripheral(self, peripheral_name):
        """Toggles the state of a specific peripheral between `on` and `off`."""
        self.state.print_status(description=f"Toggling {peripheral_name}.")
        peripheral = self.info.get_resource_by_name(
            "peripherals",
            peripheral_name)
        if peripheral is None:
            return
        peripheral_id = peripheral["id"]
        toggle_peripheral_message = {
            "kind": "toggle_pin",
            "args": {
                "pin_number": {
                    "kind": "named_pin",
                    "args": {
                        "pin_type": "Peripheral",
                        "pin_id": peripheral_id,
                    }
                }
            }
        }

        self.broker.publish(toggle_peripheral_message)

    def on(self, pin_number):
        """Turns specified pin number `on` (100%)."""
        self.state.print_status(
            description=f"Turning ON pin number {pin_number}.")

        on_message = {
            "kind": "write_pin",
            "args": {
                "pin_value": 1,
                "pin_mode": 0,
                "pin_number": pin_number,
            }
        }

        self.broker.publish(on_message)

    def off(self, pin_number):
        """Turns specified pin number `off` (0%)."""
        self.state.print_status(
            description=f"Turning OFF pin number {pin_number}.")

        off_message = {
            "kind": "write_pin",
            "args": {
                "pin_value": 0,
                "pin_mode": 0,
                "pin_number": pin_number,
            }
        }

        self.broker.publish(off_message)



"""
JobHandling class.
"""

# └── functions/jobs.py
#     ├── [BROKER] get_job()
#     ├── [BROKER] set_job()
#     └── [BROKER] complete_job()

from .broker import BrokerConnect

from .information import Information
from .resources import Resources


class JobHandling():
    """Job handling class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.info = Information(state)
        self.resource = Resources(state)
        self.state = state

    def get_job(self, job_name=None):
        """Retrieves the status or details of the specified job."""
        self.state.print_status(description="Retrieving job data...")

        status_data = self.info.read_status()

        if status_data is None:
            error = "ERROR: No job data available."
            self.state.print_status(description=error, update_only=True)
            self.state.error = error
            return

        if job_name is None:
            jobs = status_data["jobs"]
        else:
            jobs = status_data["jobs"][job_name]

        self.state.print_status(endpoint_json=jobs, update_only=True)
        return jobs

    def set_job(self, job_name, status, percent):
        """Initiates or modifies job with given parameters."""
        self.state.print_status(
            description=f"Marking job {job_name} as {percent}% {status}.")

        lua_code = f"""
            local job_name = "{job_name}"
            set_job(job_name)

            -- Update the job\'s status and percent:
            set_job(job_name, {{
            status = "{status}",
            percent = {percent}
            }})
        """

        self.resource.lua(lua_code)

    def complete_job(self, job_name):
        """Marks job as completed and triggers any associated actions."""
        self.state.print_status(
            description=f"Marking job {job_name} as `complete`.")

        lua_code = f"complete_job(\"{job_name}\")"

        self.resource.lua(lua_code)



"""
BasicCommands class.
"""

# └── functions/basic_commands.py
#     ├── [BROKER] wait()
#     ├── [BROKER] e_stop()
#     ├── [BROKER] unlock()
#     ├── [BROKER] reboot()
#     └── [BROKER] shutdown()

from .broker import BrokerConnect


class BasicCommands():
    """Basic commands class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.state = state

    def wait(self, duration):
        """Pauses execution for a certain number of milliseconds."""

        description = f"Waiting for {duration} milliseconds..."
        self.state.print_status(description=description)

        wait_message = {
            "kind": "wait",
            "args": {
                "milliseconds": duration
            }
        }

        self.broker.publish(wait_message)

    def e_stop(self):
        """Emergency locks (E-stops) the Farmduino microcontroller."""

        self.state.print_status(description="Emergency stopping device")

        stop_message = {
            "kind": "emergency_lock",
            "args": {}
        }

        stop_message = self.broker.wrap_message(stop_message, priority=9000)
        self.broker.publish(stop_message)

    def unlock(self):
        """Unlocks a locked (E-stopped) device."""

        self.state.print_status(description="Unlocking device")

        unlock_message = {
            "kind": "emergency_unlock",
            "args": {}
        }

        unlock_message = self.broker.wrap_message(
            unlock_message,
            priority=9000)
        self.broker.publish(unlock_message)

    def reboot(self):
        """Reboots the FarmBot OS and re-initializes the device."""

        self.state.print_status(description="Rebooting device")

        reboot_message = {
            "kind": "reboot",
            "args": {
                "package": "farmbot_os"
            }
        }

        self.broker.publish(reboot_message)

    def shutdown(self):
        """Shuts down the FarmBot OS and turns the device off."""

        self.state.print_status(description="Shutting down device")

        shutdown_message = {
            "kind": "power_off",
            "args": {}
        }

        self.broker.publish(shutdown_message)



"""
Camera class.
"""

# └── functions/camera.py
#     ├── [BROKER] calibrate_camera()
#     ├── [BROKER] take_photo()
#     └── [BROKER] photo_grid()

from .broker import BrokerConnect


class Camera():
    """Camera class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.state = state

    def calibrate_camera(self):
        """Performs camera calibration. This action will reset camera calibration settings."""

        self.state.print_status(description="Calibrating camera")

        calibrate_message = {
            "kind": "execute_script",
            "args": {
                "label": "camera-calibration"
            },
        }

        self.broker.publish(calibrate_message)

    def take_photo(self):
        """Takes photo using the device camera and uploads it to the web app."""

        self.state.print_status(description="Taking a photo")

        photo_message = {
            "kind": "take_photo",
            "args": {}
        }

        self.broker.publish(photo_message)

    # TODO: photo_grid()



"""
ApiConnect class.
"""

# └── functions/api.py
#     ├── [API] get_token()
#     ├── [API] check_token()
#     ├── [API] request_handling()
#     └── [API] request()

import json
from html.parser import HTMLParser
import requests


class HTMLResponseParser(HTMLParser):
    """Response parser for HTML content."""

    def __init__(self):
        super().__init__()
        self.is_header = False
        self.headers = []

    def read(self, data):
        """Read the headers from the HTML content."""
        self.is_header = False
        self.headers = []
        self.reset()
        self.feed(data)
        return " ".join(self.headers)

    def handle_starttag(self, tag, attrs):
        """Detect headers."""
        if tag in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
            self.is_header = True

    def handle_data(self, data):
        """Add header data to the list."""
        if self.is_header:
            self.headers.append(data.strip())
            self.is_header = False


class ApiConnect():
    """Connect class for FarmBot API."""

    def __init__(self, state):
        self.state = state

    def _request(self, **kwargs):
        """Internal method to make requests."""
        response = None
        try:
            response = requests.request(
                method=kwargs["method"],
                url=kwargs["url"],
                headers=kwargs["headers"],
                json=kwargs["json"],
                timeout=kwargs["timeout"])
        except requests.exceptions.RequestException as e:
            if isinstance(e, requests.exceptions.SSLError):
                self.state.error = "ERROR: The server does not support SSL."
                self.state.error += f' ({self.state.ssl=})'
            elif isinstance(e, requests.exceptions.ConnectionError):
                self.state.error = "ERROR: The server address does not exist."
            elif isinstance(e, requests.exceptions.Timeout):
                self.state.error = "ERROR: The request timed out."
            elif isinstance(e, requests.exceptions.RequestException):
                self.state.error = "ERROR: There was a problem with the request."
        except Exception as e:
            self.state.error = f"ERROR: An unexpected error occurred: {e}"
        return response

    def get_token(self, email, password, server="https://my.farm.bot"):
        """Get FarmBot authorization token. Server is 'https://my.farm.bot' by default."""
        self.state.ssl = "https" in server
        headers = {'content-type': 'application/json'}
        user = {'user': {'email': email, 'password': password}}
        timeout = self.state.timeout["api"]
        response = self._request(
            method='POST',
            url=f'{server}/api/tokens',
            headers=headers,
            json=user,
            timeout=timeout)
        if response is not None:
            if response.status_code == 200:
                self.state.token = response.json()
                self.state.error = None
                description = f"Successfully fetched token from {server}."
                self.state.print_status(description=description)
                return response.json()
            if response.status_code == 404:
                self.state.error = "HTTP ERROR: The server address does not exist."
            elif response.status_code == 422:
                self.state.error = "HTTP ERROR: Incorrect email address or password."
            else:
                code = response.status_code
                self.state.error = f"HTTP ERROR: Unexpected status code {code}"

        self.state.token = None
        self.state.print_status(description=self.state.error)
        return self.state.error

    @staticmethod
    def parse_text(text):
        """Parse response text."""
        if '<html' in text:
            parser = HTMLResponseParser()
            return parser.read(text)
        return text

    def request_handling(self, response, make_request):
        """Handle errors associated with different endpoint errors."""

        error_messages = {
            404: "The specified endpoint does not exist.",
            400: "The specified ID is invalid or you do not have access to it.",
            401: "The user`s token has expired or is invalid.",
            502: "Please check your internet connection and try again."
        }

        text = self.parse_text(response.text)

        # Handle HTTP status codes
        if response.status_code == 200:
            if not make_request:
                description = "Editing disabled, request not sent."
            else:
                description = "Successfully sent request via API."
            self.state.print_status(description=description)
            return 200
        if 400 <= response.status_code < 500:
            err = error_messages.get(response.status_code, response.reason)
            self.state.error = f"CLIENT ERROR {response.status_code}: {err}"
        elif 500 <= response.status_code < 600:
            self.state.error = f"SERVER ERROR {response.status_code}: {text}"
        else:
            code = response.status_code
            self.state.error = f"UNEXPECTED ERROR {code}: {text}"

        try:
            response.json()
        except (json.JSONDecodeError, requests.exceptions.RequestException):
            self.state.error += f" ({text})"
        else:
            self.state.error += f" ({json.dumps(response.json(), indent=2)})"

        self.state.print_status(description=self.state.error)
        return response.status_code

    def request(self, method, endpoint, database_id, payload=None):
        """Make requests to API endpoints using different methods."""

        self.state.check_token()

        # use 'GET' method to view endpoint data
        # use 'POST' method to overwrite/create new endpoint data
        # use 'PATCH' method to edit endpoint data (used for new logs)
        # use 'DELETE' method to delete endpoint data

        token = self.state.token["token"]
        iss = token["unencoded"]["iss"]

        id_part = "" if database_id is None else f"/{database_id}"
        http_part = "https" if self.state.ssl else "http"
        url = f'{http_part}:{iss}/api/{endpoint}{id_part}'

        headers = {'authorization': token['encoded'],
                   'content-type': 'application/json'}
        make_request = not self.state.dry_run or method == "GET"
        if make_request:
            timeout = self.state.timeout["api"]
            response = self._request(
                method=method,
                url=url,
                headers=headers,
                json=payload,
                timeout=timeout)
        else:
            response = requests.Response()
            response.status_code = 200
            response._content = b'{"edit_requests_disabled": true}'

        if response is not None and self.request_handling(response, make_request) == 200:
            self.state.error = None
            description = "Successfully fetched request contents."
            self.state.print_status(description=description)
            return response.json()
        description = "There was an error processing the request..."
        self.state.print_status(description=description)
        return self.state.error
